import re
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

def bigplot(Z,Ldf,Ydf,Ndf,LdfMX,LdfMN,SHIFT=0,MXMN=False,save=False):
    L_ChSt = list(set(Ldf.loc[Z,'Z']-Ldf.loc[Z,'q']))
    L_ChSt.reverse()
    Y_ChSt = list(set(list(Ydf.loc[Z,'Z-q'])))
    Y_ChSt.reverse()
    
    if MXMN:
        Lmx_ChSt = list(set(LdfMX.loc[Z,'Z']-LdfMX.loc[Z,'q']))
        Lmx_ChSt.reverse()
        Lmn_ChSt = list(set(LdfMN.loc[Z,'Z']-LdfMN.loc[Z,'q']))
        Lmn_ChSt.reverse()

    # Plot from Experiment       
    for q in set(Ydf.loc[Z]['q']):
        if q == min(Ydf.loc[Z]['q']):
            plt.plot(Ydf.loc[(Z,q),'A'],Ydf.loc[(Z,q),'Yield'],
                     color='tab:red', linewidth=3, marker='o', label=f'Experiment: {Y_ChSt} (L to R)')
        else:
            plt.plot(Ydf.loc[(Z,q),'A'],Ydf.loc[(Z,q),'Yield'],
                     color='tab:red', linewidth=3, marker='o')

    # Plot from LISE
    for q in set(Ldf.loc[Z]['q']):
        if q == min(Ldf.loc[Z]['q']):
            plt.plot(Ldf.loc[(Z,q),'A'],Ldf.loc[(Z,q),'Rate'],
                     color='tab:orange', marker='s', label=f'LISE Original: {L_ChSt} (L to R)')
        else:
            plt.plot(Ldf.loc[(Z,q),'A'],Ldf.loc[(Z,q),'Rate'],color='tab:orange',marker='s')

    # Plot of LISE shifted
    if SHIFT:
        for q in set(Ldf.loc[Z]['q']):
            NORM = float(Ndf[(Ndf['Z'] == Z) & (Ndf['q'] == q)]['Norm'])
            if q == min(Ldf.loc[Z]['q']):
                plt.plot(Ldf.loc[(Z,q),'A'],Ldf.loc[(Z,q),'Rate']/NORM,
                         color='tab:blue',marker='^', label=f'LISE Shifted: {L_ChSt} (L to R)')
            else:
                plt.plot(Ldf.loc[(Z,q),'A'],Ldf.loc[(Z,q),'Rate']/NORM,color='tab:blue',marker='^')
                
        for q in set(Ldf.loc[Z]['q']):
            UNNORM = float(Ndf[(Ndf['Z'] == Z) & (Ndf['q'] == q)]['UnNorm'])
            if q == min(Ldf.loc[Z]['q']):
                plt.plot(Ldf.loc[(Z,q),'A'],Ldf.loc[(Z,q),'Rate']/UNNORM,
                         color='tab:green',marker='x', label=f'LISE UnNormalized: {L_ChSt} (L to R)')
            else:
                plt.plot(Ldf.loc[(Z,q),'A'],Ldf.loc[(Z,q),'Rate']/UNNORM,color='tab:green',marker='x')

    # Set plot limits and create the legend
    plt.ylim(0, max(max(Ydf.loc[Z, 'Yield']), max(Ldf.loc[Z, 'Rate'])))
    plt.legend()
    
#     for q in set(Ydf.loc[Z].index):
#         plt.plot(list(Ydf.loc[(Z,q),'A']),list(Ydf.loc[(Z,q),'Yield']),color='tab:red',linewidth=3,marker='o')
#     #     print(list(Ydf.loc[(Z,q),'A']),list(Ydf.loc[(Z,q),'Norm']))
#     for q in set(Ldf.loc[Z].index):
#         plt.plot(list(Ldf.loc[(Z,q),'A']),list(Ldf.loc[(Z,q),'Rate']),color='tab:orange',marker='s')
#     #     print(list(Ldf.loc[(Z,q),'A']),list(Ldf.loc[(Z,q),'Norm']))


    ylim = max([max(list(Ldf.loc[Z,'Rate'])),max(list(Ydf.loc[Z,'Yield']))])
#     lgd = [f'Experiment: {Y_ChSt}(L to R)']

#     for i in range(len(Y_ChSt)-1):
#         lgd.append('_Hidden')
#     lgd.append(f'LISE States: {L_ChSt}(L to R)')
    

    # Plot of Max Mins
    if MXMN == True:
        for q in set(LdfMX.loc[Z]['q']):
            if q == min(LdfMX.loc[Z]['q']):
                plt.plot(list(LdfMX.loc[(Z,q),'A']),list(LdfMX.loc[(Z,q),'Rate']),
                         color='tab:green', label=f'LISE MAX: {L_ChSt}(L to R)', marker='s')
            else:
                plt.plot(list(LdfMX.loc[(Z,q),'A']),list(LdfMX.loc[(Z,q),'Rate']),color='tab:green',marker='s')
        for q in set(LdfMN.loc[Z]['q']):
            if q == min(LdfMN.loc[Z]['q']):
                plt.plot(list(LdfMN.loc[(Z,q),'A']),list(LdfMN.loc[(Z,q),'Rate']),
                         color='tab:purple', label=f'LISE MIN: {L_ChSt}(L to R)', marker='s')
            else:
                plt.plot(list(LdfMN.loc[(Z,q),'A']),list(LdfMN.loc[(Z,q),'Rate']),color='tab:purple',marker='s') 

#         for i in range(len(L_ChSt)-1):
#             lgd.append('_Hidden')
#         lgd.append(f'LISE MAX: {L_ChSt}(L to R)')

#         for i in range(len(Lmx_ChSt)-1):
#             lgd.append('_Hidden')
#         lgd.append(f'LISE MIN: {L_ChSt}(L to R)')
        

    plt.yscale('log')
    plt.xlabel('Mass Number (A)')
    plt.ylabel('Total Count (#)')
    plt.ylim(5e-2,ylim*100000)
    plt.legend(loc='upper center')
    plt.title(f'Z{Z}')
    
def obj(x,*args):
    return logchi(args[0], args[1], WTD=args[2], WT=args[3], norm=x)

def logchi(datfix,datmov,DEBUG=False,norm=1,WTD=False,WT=[]):
    """Average log difference"""
    
    if DEBUG:
        print(f"In func logchi. Vars norm,WTD = ({norm},{WTD})")
        print(f"Weights: {WT}")
    
    # @@@ Careful about inplace modifications!!!
    logfix = datfix.copy(deep=True)
    logmov = datmov.copy(deep=True)

    logmov *= norm #A global kind of normalization, to help match shape over proximity to epax

    logfix = np.log10(logfix)
    logmov = np.log10(logmov)
    chi = np.abs(logfix-logmov)

    if WTD:
        return sum(chi*WT)/sum(WT)
    else:
        return sum(chi)/len(chi)
        
def read_res(LISEfile):
    """Prep LISE res file as easy to use MultiIndex DataFrame"""
    Ldf = pd.read_csv(LISEfile,sep='\t'); Ldf = Ldf.drop(index=0)
    Ldf.columns = ['A','Z','q','Reaction','Rate','Trans','XS','q[L]-trans','qAll-trans']
    Ldf['Rate'] = Ldf['Rate'].astype(float)
    Ldf['A'] = Ldf['A'].astype(int)
    Ldf['Z'] = Ldf['Z'].astype(int)
    Ldf['q'] = Ldf['q'].astype(int)
    Ldf['Trans'] = Ldf['Trans'].astype(float)
    Ldf['ID'] = Ldf.Z*100000+Ldf.A*100+Ldf.q

    tuples = list(zip(list(Ldf['Z']),list(Ldf['q']),list(Ldf['A'])))
    index = pd.MultiIndex.from_tuples(tuples, names=["Z","q","A"])
    # Ldf['Norm'] = Ldf['Rate']/max(Ldf['Rate'])
    Ldf.index = index
    return Ldf.sort_index()

def read_yield(YIELDfile):
    """Prep Excel Yields with easy to use MultiIndex"""
    Ydf = pd.read_excel(YIELDfile,sheet_name='base')
#     Ydf = Ydf.drop_duplicates()
    for i in Ydf.index:
        if Ydf.loc[i,'Yield'] < 5:
            Ydf = Ydf.drop(index=i)
    Ydf.rename(columns={'Unnamed: 0': 'ID'}, inplace=True)

    tuples = list(zip(list(Ydf['Z']),list(Ydf['q']),list(Ydf['A'])))
    index = pd.MultiIndex.from_tuples(tuples, names=["Z","q","A"])
    # Ydf['Norm'] = Ydf['Yield']/max(Ydf['Yield'])
    Ydf.index = index
    return Ydf.sort_index()
    
def read_res1(file):
    """Read the full transmission output files from LISE, extension res1"""
    # Read file
    with open(file, 'r') as file:
        lines = file.readlines()

    # Search for the beginning of the transmission table
    for i,L in enumerate(lines):
        if '| A El| A | Z |' in L:
            titlerow = i

    # Build the column titles for the DataFrame
    li1 = lines[titlerow].split('|')
    li2 = lines[titlerow+1].split('|')
    col_title = [a.replace(' ','') + ' ' + b.replace(' ','') for a, b in zip(li1, li2)]
    col_title = [c.strip(' ') for c in col_title]

    # Build the DataFrame
    trans = lines[titlerow+2:-4]
    trans = list(filter(lambda x: '--' not in x, trans))
    NesLi=[]
    for ROW in trans:
        NesLi.append([cell.replace(' ','') for cell in ROW.split('|')])
    # Loop needed for nested list (NesLi) operation. Cleaner that double list comp.
    df = pd.DataFrame(NesLi,columns=col_title)

    # Assign correct data types
    string_columns = ['AEl', 'Re','','\n \n']
    integer_columns = ['A', 'Z', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7']
    float_columns = [c for c in col_title if 'trans' in c]

    dtype_dict = {col: 'str' for col in string_columns}
    dtype_dict.update({col: 'int64' for col in integer_columns})
    dtype_dict.update({col: 'float64' for col in float_columns})
    df = df.astype(dtype_dict)

    # Build the unique index
    q = re.compile('Q\d')
    qLi = [q.findall(c) for c in col_title]
    # Extract Q values from non-empty entries and find the highest Q --- ChatGPT
    Q = max(['Q0'] + [entry[0] for entry in qLi if entry], key=lambda x: int(x[1:]))
    df.insert(4, 'Qf', df[Q])
    df.insert(0, 'ID', df.Z*100000+df.A*100+df.Qf)
    return df